#' @title Estimate differential causal effect using two-stage least squares.
#' 
#' @description
#' 
#' This method is the simplest variation of two-stage least squares.
#' 
#' @param W         array containing indices of the instrumental variables.
#' @param Z         array containing indices of the conditioning variables.
#' @param x         index of treatment variable.
#' @param y         index of outcome variable.
#' @param cov_model covariance matrix of all variables
#'        
#' @return Differential causal effect as implied by \code{cov_model}.

IV_TSLS <- function(W, Z, x, y, cov_model)
{
  WZ <- c(W, Z)
  XZ <- c(x, Z)
  
  # First stage
  
  beta_x <- solve(cov_model[WZ, WZ], cov_model[WZ, x])
  
  # Build covariance of X, Z, R and Y
  
  cov_XZ.R <- cov_model[c(x, Z), x] - colSums(beta_x * cov_model[WZ, XZ, drop = FALSE])
  cov_RR   <- cov_model[x, x] - sum(cov_model[x, WZ] * beta_x)
  cov_Y.R  <- cov_model[y, x] - sum(beta_x * cov_model[y, WZ])
  
  cov_XZR   <- rbind(cbind(cov_model[XZ, XZ], cov_XZ.R), c(cov_XZ.R, cov_RR))
  cov_XZR.Y <- c(cov_model[XZ, y], cov_Y.R)  
  
  # Second stage
  
  beta_y <- solve(cov_XZR, cov_XZR.Y)
  
  # Return
  
  return(as.numeric(beta_y[1]))
}

standardize_data <- function(dat) 
{
  N <- nrow(dat)
  sd_dat <- apply(dat, 2, sd)
  m_dat <- colMeans(dat)
  dat <- (dat - t(matrix(rep(m_dat, N), nrow = length(m_dat)))) / t(matrix(rep(sd_dat, N), nrow = length(sd_dat)))
  return(dat) 
}
